@@ -1,10 +1,10 @@
-
-from electrum.i18n import _
-from electrum.plugin import run_hook
 from PyQt5.QtGui import *
 from PyQt5.QtCore import *
 from PyQt5.QtWidgets import QFileDialog
 
+from electrum.i18n import _
+from electrum.plugin import run_hook
+
 from .util import ButtonsTextEdit, MessageBoxMixin, ColorScheme
 
 
@@ -13,7 +13,8 @@ class ShowQRTextEdit(ButtonsTextEdit):
     def __init__(self, text=None):
         ButtonsTextEdit.__init__(self, text)
         self.setReadOnly(1)
-        self.addButton(":icons/qrcode.png", self.qr_show, _("Show as QR code"))
+        icon = ":icons/qrcode_white.png" if ColorScheme.dark_scheme else ":icons/qrcode.png"
+        self.addButton(icon, self.qr_show, _("Show as QR code"))
 
         run_hook('show_text_edit', self)
 
@@ -38,7 +39,7 @@ class ScanQRTextEdit(ButtonsTextEdit, MessageBoxMixin):
         self.allow_multi = allow_multi
         self.setReadOnly(0)
         self.addButton(":icons/file.png", self.file_input, _("Read file"))
-        icon = ":icons/qrcode_white.png" if ColorScheme.dark_scheme else ":icons/qrcode.png"
+        icon = ":icons/camera_white.png" if ColorScheme.dark_scheme else ":icons/camera_dark.png"
         self.addButton(icon, self.qr_input, _("Read QR code"))
         run_hook('scan_text_edit', self)
 
