@@ -23,41 +23,40 @@
 # CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
 
+from PyQt5.QtGui import QStandardItemModel, QStandardItem
+from PyQt5.QtWidgets import QMenu
+from PyQt5.QtCore import Qt
+
 from electrum.i18n import _
 from electrum.util import format_time, age
 from electrum.plugin import run_hook
 from electrum.paymentrequest import PR_UNKNOWN
-from PyQt5.QtGui import *
-from PyQt5.QtCore import *
-from PyQt5.QtWidgets import QTreeWidgetItem, QMenu
-from .util import MyTreeWidget, pr_tooltips, pr_icons
+from electrum.wallet import InternalAddressCorruption
 
+from .util import MyTreeView, pr_tooltips, pr_icons
 
-class RequestList(MyTreeWidget):
+class RequestList(MyTreeView):
     filter_columns = [0, 1, 2, 3, 4]  # Date, Account, Address, Description, Amount
 
 
     def __init__(self, parent):
-        MyTreeWidget.__init__(self, parent, self.create_menu, [_('Date'), _('Address'), '', _('Description'), _('Amount'), _('Status')], 3)
-        self.currentItemChanged.connect(self.item_changed)
-        self.itemClicked.connect(self.item_changed)
+        super().__init__(parent, self.create_menu, 3, editable_columns=[])
+        self.setModel(QStandardItemModel(self))
         self.setSortingEnabled(True)
         self.setColumnWidth(0, 180)
-        self.hideColumn(1)
+        self.update()
+        self.selectionModel().currentRowChanged.connect(self.item_changed)
 
-    def item_changed(self, item):
-        if item is None:
-            return
-        if not item.isSelected():
-            return
-        addr = str(item.text(1))
+    def item_changed(self, idx):
+        # TODO use siblingAtColumn when min Qt version is >=5.11
+        addr = self.model().itemFromIndex(idx.sibling(idx.row(), 1)).text()
         req = self.wallet.receive_requests.get(addr)
         if req is None:
             self.update()
             return
         expires = age(req['time'] + req['exp']) if req.get('exp') else _('Never')
         amount = req['amount']
-        message = self.wallet.labels.get(addr, '')
+        message = req['memo']
         self.parent.receive_address_e.setText(addr)
         self.parent.receive_message_e.setText(message)
         self.parent.receive_amount_e.setAmount(amount)
@@ -66,26 +65,32 @@ class RequestList(MyTreeWidget):
         self.parent.expires_label.setText(expires)
         self.parent.new_request_button.setEnabled(True)
 
-    def on_update(self):
+    def update(self):
         self.wallet = self.parent.wallet
         # hide receive tab if no receive requests available
-        b = len(self.wallet.receive_requests) > 0
-        self.setVisible(b)
-        self.parent.receive_requests_label.setVisible(b)
-        if not b:
-            self.parent.expires_label.hide()
-            self.parent.expires_combo.show()
+        if self.parent.isVisible():
+            b = len(self.wallet.receive_requests) > 0
+            self.setVisible(b)
+            self.parent.receive_requests_label.setVisible(b)
+            if not b:
+                self.parent.expires_label.hide()
+                self.parent.expires_combo.show()
 
         # update the receive address if necessary
         current_address = self.parent.receive_address_e.text()
         domain = self.wallet.get_receiving_addresses()
-        addr = self.wallet.get_unused_address()
+        try:
+            addr = self.wallet.get_unused_address()
+        except InternalAddressCorruption as e:
+            self.parent.show_error(str(e))
+            addr = ''
         if not current_address in domain and addr:
             self.parent.set_receive_address(addr)
         self.parent.new_request_button.setEnabled(addr != current_address)
 
-        # clear the list and fill it again
-        self.clear()
+        self.model().clear()
+        self.update_headers([_('Date'), _('Address'), '', _('Description'), _('Amount'), _('Status')])
+        self.hideColumn(1) # hide address column
         for req in self.wallet.get_sorted_requests(self.config):
             address = req['address']
             if address not in domain:
@@ -93,35 +98,40 @@ class RequestList(MyTreeWidget):
             timestamp = req.get('time', 0)
             amount = req.get('amount')
             expiration = req.get('exp', None)
-            message = req.get('memo', '')
+            message = req['memo']
             date = format_time(timestamp)
             status = req.get('status')
             signature = req.get('sig')
             requestor = req.get('name', '')
             amount_str = self.parent.format_amount(amount) if amount else ""
-            item = QTreeWidgetItem([date, address, '', message, amount_str, pr_tooltips.get(status,'')])
+            labels = [date, address, '', message, amount_str, pr_tooltips.get(status,'')]
+            items = [QStandardItem(e) for e in labels]
+            self.set_editability(items)
             if signature is not None:
-                item.setIcon(2, self.icon_cache.get(":icons/seal.png"))
-                item.setToolTip(2, 'signed by '+ requestor)
+                items[2].setIcon(self.icon_cache.get(":icons/seal.png"))
+                items[2].setToolTip('signed by '+ requestor)
             if status is not PR_UNKNOWN:
-                item.setIcon(6, self.icon_cache.get(pr_icons.get(status)))
-            self.addTopLevelItem(item)
-
+                items[5].setIcon(self.icon_cache.get(pr_icons.get(status)))
+            items[3].setData(address, Qt.UserRole)
+            self.model().insertRow(self.model().rowCount(), items)
 
     def create_menu(self, position):
-        item = self.itemAt(position)
+        idx = self.indexAt(position)
+        # TODO use siblingAtColumn when min Qt version is >=5.11
+        item = self.model().itemFromIndex(idx.sibling(idx.row(), 1))
         if not item:
             return
-        addr = str(item.text(1))
+        addr = item.text()
         req = self.wallet.receive_requests.get(addr)
         if req is None:
             self.update()
             return
-        column = self.currentColumn()
-        column_title = self.headerItem().text(column)
-        column_data = item.text(column)
+        column = idx.column()
+        column_title = self.model().horizontalHeaderItem(column).text()
+        column_data = item.text()
         menu = QMenu(self)
-        menu.addAction(_("Copy {}").format(column_title), lambda: self.parent.app.clipboard().setText(column_data))
+        if column != 2:
+            menu.addAction(_("Copy {}").format(column_title), lambda: self.parent.app.clipboard().setText(column_data))
         menu.addAction(_("Copy URI"), lambda: self.parent.view_and_paste('URI', '', self.parent.get_request_URI(addr)))
         menu.addAction(_("Save as BIP70 file"), lambda: self.parent.export_payment_request(addr))
         menu.addAction(_("Delete"), lambda: self.parent.delete_payment_request(addr))
