@@ -29,7 +29,7 @@ import json
 import copy
 import re
 import stat
-import hmac, hashlib
+import hashlib
 import base64
 import zlib
 from collections import defaultdict
@@ -54,7 +54,7 @@ def multisig_type(wallet_type):
     otherwise return None.'''
     if not wallet_type:
         return None
-    match = re.match('(\d+)of(\d+)', wallet_type)
+    match = re.match(r'(\d+)of(\d+)', wallet_type)
     if match:
         match = [int(x) for x in match.group(1, 2)]
     return match
@@ -73,7 +73,7 @@ class JsonDB(PrintError):
     def __init__(self, path):
         self.db_lock = threading.RLock()
         self.data = {}
-        self.path = path
+        self.path = os.path.normcase(os.path.abspath(path))
         self.modified = False
 
     def get(self, key, default=None):
@@ -90,7 +90,7 @@ class JsonDB(PrintError):
             json.dumps(key, cls=util.MyEncoder)
             json.dumps(value, cls=util.MyEncoder)
         except:
-            self.print_error("json error: cannot save", key)
+            self.print_error(f"json error: cannot save {repr(key)} ({repr(value)})")
             return
         with self.db_lock:
             if value is not None:
@@ -122,12 +122,7 @@ class JsonDB(PrintError):
             os.fsync(f.fileno())
 
         mode = os.stat(self.path).st_mode if os.path.exists(self.path) else stat.S_IREAD | stat.S_IWRITE
-        # perform atomic write on POSIX systems
-        try:
-            os.rename(temp_path, self.path)
-        except:
-            os.remove(self.path)
-            os.rename(temp_path, self.path)
+        os.replace(temp_path, self.path)
         os.chmod(self.path, mode)
         self.print_error("saved", self.path)
         self.modified = False
@@ -142,8 +137,8 @@ class JsonDB(PrintError):
 class WalletStorage(JsonDB):
 
     def __init__(self, path, manual_upgrades=False):
-        self.print_error("wallet path", path)
         JsonDB.__init__(self, path)
+        self.print_error("wallet path", path)
         self.manual_upgrades = manual_upgrades
         self.pubkey = None
         if self.file_exists():
@@ -175,6 +170,8 @@ class WalletStorage(JsonDB):
                     self.print_error('Failed to convert label to json format', key)
                     continue
                 self.data[key] = value
+        if not isinstance(self.data, dict):
+            raise WalletFileException("Malformed wallet file (not dict)")
 
         # check here if I need to load a plugin
         t = self.get('wallet_type')
@@ -575,11 +572,14 @@ class WalletStorage(JsonDB):
         # delete verified_tx3 as its structure changed
         if not self._is_upgrade_method_needed(17, 17):
             return
-
         self.put('verified_tx3', None)
-
         self.put('seed_version', 18)
 
+    # def convert_version_19(self):
+    #     TODO for "next" upgrade:
+    #       - move "pw_hash_version" from keystore to storage
+    #     pass
+
     def convert_imported(self):
         if not self._is_upgrade_method_needed(0, 13):
             return
