@@ -27,22 +27,22 @@
 
 # Note: The deserialization code originally comes from ABE.
 
-from typing import Sequence, Union
-
-from .util import print_error, profiler
-
-from . import ecc
-from . import bitcoin
-from .bitcoin import *
 import struct
 import traceback
 import sys
+from typing import (Sequence, Union, NamedTuple, Tuple, Optional, Iterable,
+                    Callable, List, Dict)
 
-#
-# Workalike python implementation of Bitcoin's CDataStream class.
-#
+from . import ecc, bitcoin, constants, segwit_addr
+from .util import print_error, profiler, to_bytes, bh2u, bfh
+from .bitcoin import (TYPE_ADDRESS, TYPE_PUBKEY, TYPE_SCRIPT, hash_160,
+                      hash160_to_p2sh, hash160_to_p2pkh, hash_to_segwit_addr,
+                      hash_encode, var_int, TOTAL_COIN_SUPPLY_LIMIT_IN_BTC, COIN,
+                      push_script, int_to_hex, push_script, b58_address_to_hash160)
+from .crypto import sha256d
 from .keystore import xpubkey_to_address, xpubkey_to_pubkey
 
+
 NO_SIGNATURE = 'ff'
 PARTIAL_TXN_HEADER_MAGIC = b'EPTF\xff'
 
@@ -59,7 +59,31 @@ class NotRecognizedRedeemScript(Exception):
     pass
 
 
+class MalformedBitcoinScript(Exception):
+    pass
+
+
+class TxOutput(NamedTuple):
+    type: int
+    address: str
+    value: Union[int, str]  # str when the output is set to max: '!'
+
+
+class TxOutputForUI(NamedTuple):
+    address: str
+    value: int
+
+
+class TxOutputHwInfo(NamedTuple):
+    address_index: Tuple
+    sorted_xpubs: Iterable[str]
+    num_sig: Optional[int]
+    script_type: str
+
+
 class BCDataStream(object):
+    """Workalike python implementation of Bitcoin's CDataStream class."""
+
     def __init__(self):
         self.input = None
         self.read_cursor = 0
@@ -253,13 +277,16 @@ def script_GetOp(_bytes : bytes):
         if opcode <= opcodes.OP_PUSHDATA4:
             nSize = opcode
             if opcode == opcodes.OP_PUSHDATA1:
-                nSize = _bytes[i]
+                try: nSize = _bytes[i]
+                except IndexError: raise MalformedBitcoinScript()
                 i += 1
             elif opcode == opcodes.OP_PUSHDATA2:
-                (nSize,) = struct.unpack_from('<H', _bytes, i)
+                try: (nSize,) = struct.unpack_from('<H', _bytes, i)
+                except struct.error: raise MalformedBitcoinScript()
                 i += 2
             elif opcode == opcodes.OP_PUSHDATA4:
-                (nSize,) = struct.unpack_from('<I', _bytes, i)
+                try: (nSize,) = struct.unpack_from('<I', _bytes, i)
+                except struct.error: raise MalformedBitcoinScript()
                 i += 4
             vch = _bytes[i:i + nSize]
             i += nSize
@@ -271,25 +298,39 @@ def script_GetOpName(opcode):
     return (opcodes.whatis(opcode)).replace("OP_", "")
 
 
-def decode_script(bytes):
-    result = ''
-    for (opcode, vch, i) in script_GetOp(bytes):
-        if len(result) > 0: result += " "
-        if opcode <= opcodes.OP_PUSHDATA4:
-            result += "%d:"%(opcode,)
-            result += short_hex(vch)
-        else:
-            result += script_GetOpName(opcode)
-    return result
+class OPPushDataGeneric:
+    def __init__(self, pushlen: Callable=None):
+        if pushlen is not None:
+            self.check_data_len = pushlen
+
+    @classmethod
+    def check_data_len(cls, datalen: int) -> bool:
+        # Opcodes below OP_PUSHDATA4 all just push data onto stack, and are equivalent.
+        return opcodes.OP_PUSHDATA4 >= datalen >= 0
+
+    @classmethod
+    def is_instance(cls, item):
+        # accept objects that are instances of this class
+        # or other classes that are subclasses
+        return isinstance(item, cls) \
+               or (isinstance(item, type) and issubclass(item, cls))
+
+
+OPPushDataPubkey = OPPushDataGeneric(lambda x: x in (33, 65))
+# note that this does not include x_pubkeys !
 
 
 def match_decoded(decoded, to_match):
+    if decoded is None:
+        return False
     if len(decoded) != len(to_match):
-        return False;
+        return False
     for i in range(len(decoded)):
-        if to_match[i] == opcodes.OP_PUSHDATA4 and decoded[i][0] <= opcodes.OP_PUSHDATA4 and decoded[i][0]>0:
-            continue  # Opcodes below OP_PUSHDATA4 all just push data onto stack, and are equivalent.
-        if to_match[i] != decoded[i][0]:
+        to_match_item = to_match[i]
+        decoded_item = decoded[i]
+        if OPPushDataGeneric.is_instance(to_match_item) and to_match_item.check_data_len(decoded_item[0]):
+            continue
+        if to_match_item != decoded_item[0]:
             return False
     return True
 
@@ -312,13 +353,13 @@ def parse_scriptSig(d, _bytes):
                     bh2u(_bytes))
         return
 
-    match = [ opcodes.OP_PUSHDATA4 ]
+    match = [OPPushDataGeneric]
     if match_decoded(decoded, match):
         item = decoded[0][1]
         if item[0] == 0:
             # segwit embedded into p2sh
             # witness version 0
-            d['address'] = bitcoin.hash160_to_p2sh(bitcoin.hash_160(item))
+            d['address'] = bitcoin.hash160_to_p2sh(hash_160(item))
             if len(item) == 22:
                 d['type'] = 'p2wpkh-p2sh'
             elif len(item) == 34:
@@ -343,7 +384,7 @@ def parse_scriptSig(d, _bytes):
     # p2pkh TxIn transactions push a signature
     # (71-73 bytes) and then their public key
     # (33 or 65 bytes) onto the stack:
-    match = [ opcodes.OP_PUSHDATA4, opcodes.OP_PUSHDATA4 ]
+    match = [OPPushDataGeneric, OPPushDataGeneric]
     if match_decoded(decoded, match):
         sig = bh2u(decoded[0][1])
         x_pubkey = bh2u(decoded[1][1])
@@ -363,7 +404,7 @@ def parse_scriptSig(d, _bytes):
         return
 
     # p2sh transaction, m of n
-    match = [ opcodes.OP_0 ] + [ opcodes.OP_PUSHDATA4 ] * (len(decoded) - 1)
+    match = [opcodes.OP_0] + [OPPushDataGeneric] * (len(decoded) - 1)
     if match_decoded(decoded, match):
         x_sig = [bh2u(x[1]) for x in decoded[1:-1]]
         redeem_script_unsanitized = decoded[-1][1]  # for partial multisig txn, this has x_pubkeys
@@ -386,7 +427,7 @@ def parse_scriptSig(d, _bytes):
         return
 
     # custom partial format for imported addresses
-    match = [ opcodes.OP_INVALIDOPCODE, opcodes.OP_0, opcodes.OP_PUSHDATA4 ]
+    match = [opcodes.OP_INVALIDOPCODE, opcodes.OP_0, OPPushDataGeneric]
     if match_decoded(decoded, match):
         x_pubkey = bh2u(decoded[2][1])
         pubkey, address = xpubkey_to_address(x_pubkey)
@@ -403,7 +444,10 @@ def parse_scriptSig(d, _bytes):
 
 
 def parse_redeemScript_multisig(redeem_script: bytes):
-    dec2 = [ x for x in script_GetOp(redeem_script) ]
+    try:
+        dec2 = [ x for x in script_GetOp(redeem_script) ]
+    except MalformedBitcoinScript:
+        raise NotRecognizedRedeemScript()
     try:
         m = dec2[0][0] - opcodes.OP_1 + 1
         n = dec2[-2][0] - opcodes.OP_1 + 1
@@ -411,7 +455,7 @@ def parse_redeemScript_multisig(redeem_script: bytes):
         raise NotRecognizedRedeemScript()
     op_m = opcodes.OP_1 + m - 1
     op_n = opcodes.OP_1 + n - 1
-    match_multisig = [ op_m ] + [opcodes.OP_PUSHDATA4]*n + [ op_n, opcodes.OP_CHECKMULTISIG ]
+    match_multisig = [op_m] + [OPPushDataGeneric] * n + [op_n, opcodes.OP_CHECKMULTISIG]
     if not match_decoded(dec2, match_multisig):
         raise NotRecognizedRedeemScript()
     x_pubkeys = [bh2u(x[1]) for x in dec2[1:-2]]
@@ -423,30 +467,36 @@ def parse_redeemScript_multisig(redeem_script: bytes):
     return m, n, x_pubkeys, pubkeys, redeem_script_sanitized
 
 
-def get_address_from_output_script(_bytes, *, net=None):
-    decoded = [x for x in script_GetOp(_bytes)]
+def get_address_from_output_script(_bytes: bytes, *, net=None) -> Tuple[int, str]:
+    try:
+        decoded = [x for x in script_GetOp(_bytes)]
+    except MalformedBitcoinScript:
+        decoded = None
 
-    # The Genesis Block, self-payments, and pay-by-IP-address payments look like:
-    # 65 BYTES:... CHECKSIG
-    match = [ opcodes.OP_PUSHDATA4, opcodes.OP_CHECKSIG ]
-    if match_decoded(decoded, match):
+    # p2pk
+    match = [OPPushDataPubkey, opcodes.OP_CHECKSIG]
+    if match_decoded(decoded, match) and ecc.ECPubkey.is_pubkey_bytes(decoded[0][1]):
         return TYPE_PUBKEY, bh2u(decoded[0][1])
 
-    # Pay-by-Bitcoin-address TxOuts look like:
-    # DUP HASH160 20 BYTES:... EQUALVERIFY CHECKSIG
-    match = [ opcodes.OP_DUP, opcodes.OP_HASH160, opcodes.OP_PUSHDATA4, opcodes.OP_EQUALVERIFY, opcodes.OP_CHECKSIG ]
+    # p2pkh
+    match = [opcodes.OP_DUP, opcodes.OP_HASH160, OPPushDataGeneric(lambda x: x == 20), opcodes.OP_EQUALVERIFY, opcodes.OP_CHECKSIG]
     if match_decoded(decoded, match):
         return TYPE_ADDRESS, hash160_to_p2pkh(decoded[2][1], net=net)
 
     # p2sh
-    match = [ opcodes.OP_HASH160, opcodes.OP_PUSHDATA4, opcodes.OP_EQUAL ]
+    match = [opcodes.OP_HASH160, OPPushDataGeneric(lambda x: x == 20), opcodes.OP_EQUAL]
     if match_decoded(decoded, match):
         return TYPE_ADDRESS, hash160_to_p2sh(decoded[1][1], net=net)
 
-    # segwit address
-    possible_witness_versions = [opcodes.OP_0] + list(range(opcodes.OP_1, opcodes.OP_16 + 1))
-    for witver, opcode in enumerate(possible_witness_versions):
-        match = [ opcode, opcodes.OP_PUSHDATA4 ]
+    # segwit address (version 0)
+    match = [opcodes.OP_0, OPPushDataGeneric(lambda x: x in (20, 32))]
+    if match_decoded(decoded, match):
+        return TYPE_ADDRESS, hash_to_segwit_addr(decoded[1][1], witver=0, net=net)
+
+    # segwit address (version 1-16)
+    future_witness_versions = list(range(opcodes.OP_1, opcodes.OP_16 + 1))
+    for witver, opcode in enumerate(future_witness_versions, start=1):
+        match = [opcode, OPPushDataGeneric(lambda x: 2 <= x <= 40)]
         if match_decoded(decoded, match):
             return TYPE_ADDRESS, hash_to_segwit_addr(decoded[1][1], witver=witver, net=net)
 
@@ -599,15 +649,12 @@ def deserialize(raw: str, force_full_parse=False) -> dict:
 
 # pay & redeem scripts
 
-
-
 def multisig_script(public_keys: Sequence[str], m: int) -> str:
     n = len(public_keys)
-    assert n <= 15
-    assert m <= n
-    op_m = format(opcodes.OP_1 + m - 1, 'x')
-    op_n = format(opcodes.OP_1 + n - 1, 'x')
-    keylist = [op_push(len(k)//2) + k for k in public_keys]
+    assert 1 <= m <= n <= 15, f'm {m}, n {n}'
+    op_m = bh2u(bytes([opcodes.OP_1 - 1 + m]))
+    op_n = bh2u(bytes([opcodes.OP_1 - 1 + n]))
+    keylist = [push_script(k) for k in public_keys]
     return op_m + ''.join(keylist) + op_n + 'ae'
 
 
@@ -630,13 +677,14 @@ class Transaction:
         else:
             raise Exception("cannot initialize transaction", raw)
         self._inputs = None
-        self._outputs = None
+        self._outputs = None  # type: List[TxOutput]
         self.locktime = 0
         self.version = 1
         # by default we assume this is a partial txn;
         # this value will get properly set when deserializing
         self.is_partial_originally = True
         self._segwit_ser = None  # None means "don't know"
+        self.output_info = None  # type: Optional[Dict[str, TxOutputHwInfo]]
 
     def update(self, raw):
         self.raw = raw
@@ -648,7 +696,7 @@ class Transaction:
             self.deserialize()
         return self._inputs
 
-    def outputs(self):
+    def outputs(self) -> List[TxOutput]:
         if self._outputs is None:
             self.deserialize()
         return self._outputs
@@ -672,7 +720,7 @@ class Transaction:
 
         `signatures` is expected to be a list of sigs with signatures[i]
         intended for self._inputs[i].
-        This is used by the Trezor and KeepKey plugins.
+        This is used by the Trezor, KeepKey an Safe-T plugins.
         """
         if self.is_complete():
             return
@@ -683,7 +731,7 @@ class Transaction:
             sig = signatures[i]
             if sig in txin.get('signatures'):
                 continue
-            pre_hash = Hash(bfh(self.serialize_preimage(i)))
+            pre_hash = sha256d(bfh(self.serialize_preimage(i)))
             sig_string = ecc.sig_string_from_der_sig(bfh(sig[:-2]))
             for recid in range(4):
                 try:
@@ -701,7 +749,6 @@ class Transaction:
                     j = pubkeys.index(pubkey_hex)
                     print_error("adding sig", i, j, pubkey_hex, sig)
                     self.add_signature_to_txin(i, j, sig)
-                    #self._inputs[i]['x_pubkeys'][j] = pubkey
                     break
         # redo raw
         self.raw = self.serialize()
@@ -713,6 +760,17 @@ class Transaction:
         txin['witness'] = None    # force re-serialization
         self.raw = None
 
+    def add_inputs_info(self, wallet):
+        if self.is_complete():
+            return
+        for txin in self.inputs():
+            wallet.add_input_info(txin)
+
+    def remove_signatures(self):
+        for txin in self.inputs():
+            txin['signatures'] = [None] * len(txin['signatures'])
+        assert not self.is_complete()
+
     def deserialize(self, force_full_parse=False):
         if self.raw is None:
             return
@@ -721,7 +779,7 @@ class Transaction:
             return
         d = deserialize(self.raw, force_full_parse)
         self._inputs = d['inputs']
-        self._outputs = [(x['type'], x['address'], x['value']) for x in d['outputs']]
+        self._outputs = [TxOutput(x['type'], x['address'], x['value']) for x in d['outputs']]
         self.locktime = d['lockTime']
         self.version = d['version']
         self.is_partial_originally = d['partial']
@@ -734,10 +792,12 @@ class Transaction:
         self._inputs = inputs
         self._outputs = outputs
         self.locktime = locktime
+        self.BIP69_sort()
         return self
 
     @classmethod
-    def pay_script(self, output_type, addr):
+    def pay_script(self, output_type, addr: str) -> str:
+        """Returns scriptPubKey in hex form."""
         if output_type == TYPE_SCRIPT:
             return addr
         elif output_type == TYPE_ADDRESS:
@@ -801,7 +861,7 @@ class Transaction:
     @classmethod
     def serialize_witness(self, txin, estimate_size=False):
         _type = txin['type']
-        if not self.is_segwit_input(txin) and not self.is_input_value_needed(txin):
+        if not self.is_segwit_input(txin) and not txin['type'] == 'address':
             return '00'
         if _type == 'coinbase':
             return txin['witness']
@@ -839,10 +899,6 @@ class Transaction:
     def is_segwit_inputtype(cls, txin_type):
         return txin_type in ('p2wpkh', 'p2wpkh-p2sh', 'p2wsh', 'p2wsh-p2sh')
 
-    @classmethod
-    def is_input_value_needed(cls, txin):
-        return cls.is_segwit_input(txin) or txin['type'] == 'address'
-
     @classmethod
     def guess_txintype_from_address(cls, addr):
         # It's not possible to tell the script type in general
@@ -857,7 +913,7 @@ class Transaction:
         witver, witprog = segwit_addr.decode(constants.net.SEGWIT_HRP, addr)
         if witprog is not None:
             return 'p2wpkh'
-        addrtype, hash_160 = b58_address_to_hash160(addr)
+        addrtype, hash_160_ = b58_address_to_hash160(addr)
         if addrtype == constants.net.ADDRTYPE_P2PKH:
             return 'p2pkh'
         elif addrtype == constants.net.ADDRTYPE_P2SH:
@@ -933,7 +989,7 @@ class Transaction:
             return multisig_script(pubkeys, txin['num_sig'])
         elif txin['type'] in ['p2wpkh', 'p2wpkh-p2sh']:
             pubkey = pubkeys[0]
-            pkh = bh2u(bitcoin.hash_160(bfh(pubkey)))
+            pkh = bh2u(hash_160(bfh(pubkey)))
             return '76a9' + push_script(pkh) + '88ac'
         elif txin['type'] == 'p2pk':
             pubkey = pubkeys[0]
@@ -968,15 +1024,16 @@ class Transaction:
         for txin in self.inputs():
             txin['sequence'] = nSequence
 
-    def BIP_LI01_sort(self):
-        # See https://github.com/kristovatlas/rfc/blob/master/bips/bip-li01.mediawiki
-        self._inputs.sort(key = lambda i: (i['prevout_hash'], i['prevout_n']))
-        self._outputs.sort(key = lambda o: (o[2], self.pay_script(o[0], o[1])))
+    def BIP69_sort(self, inputs=True, outputs=True):
+        if inputs:
+            self._inputs.sort(key = lambda i: (i['prevout_hash'], i['prevout_n']))
+        if outputs:
+            self._outputs.sort(key = lambda o: (o.value, self.pay_script(o.type, o.address)))
 
-    def serialize_output(self, output):
-        output_type, addr, amount = output
-        s = int_to_hex(amount, 8)
-        script = self.pay_script(output_type, addr)
+    @classmethod
+    def serialize_output(cls, output: TxOutput) -> str:
+        s = int_to_hex(output.value, 8)
+        script = cls.pay_script(output.type, output.address)
         s += var_int(len(script)//2)
         s += script
         return s
@@ -990,9 +1047,9 @@ class Transaction:
         txin = inputs[i]
         # TODO: py3 hex
         if self.is_segwit_input(txin):
-            hashPrevouts = bh2u(Hash(bfh(''.join(self.serialize_outpoint(txin) for txin in inputs))))
-            hashSequence = bh2u(Hash(bfh(''.join(int_to_hex(txin.get('sequence', 0xffffffff - 1), 4) for txin in inputs))))
-            hashOutputs = bh2u(Hash(bfh(''.join(self.serialize_output(o) for o in outputs))))
+            hashPrevouts = bh2u(sha256d(bfh(''.join(self.serialize_outpoint(txin) for txin in inputs))))
+            hashSequence = bh2u(sha256d(bfh(''.join(int_to_hex(txin.get('sequence', 0xffffffff - 1), 4) for txin in inputs))))
+            hashOutputs = bh2u(sha256d(bfh(''.join(self.serialize_output(o) for o in outputs))))
             outpoint = self.serialize_outpoint(txin)
             preimage_script = self.get_preimage_script(txin)
             scriptCode = var_int(len(preimage_script) // 2) + preimage_script
@@ -1021,6 +1078,7 @@ class Transaction:
             return network_ser
 
     def serialize_to_network(self, estimate_size=False, witness=True):
+        self.deserialize()
         nVersion = int_to_hex(self.version, 4)
         nLocktime = int_to_hex(self.locktime, 4)
         inputs = self.inputs()
@@ -1045,22 +1103,24 @@ class Transaction:
         if not all_segwit and not self.is_complete():
             return None
         ser = self.serialize_to_network(witness=False)
-        return bh2u(Hash(bfh(ser))[::-1])
+        return bh2u(sha256d(bfh(ser))[::-1])
 
     def wtxid(self):
         self.deserialize()
         if not self.is_complete():
             return None
         ser = self.serialize_to_network(witness=True)
-        return bh2u(Hash(bfh(ser))[::-1])
+        return bh2u(sha256d(bfh(ser))[::-1])
 
     def add_inputs(self, inputs):
         self._inputs.extend(inputs)
         self.raw = None
+        self.BIP69_sort(outputs=False)
 
     def add_outputs(self, outputs):
         self._outputs.extend(outputs)
         self.raw = None
+        self.BIP69_sort(inputs=False)
 
     def input_value(self):
         return sum(x['value'] for x in self.inputs())
@@ -1074,7 +1134,6 @@ class Transaction:
     def is_final(self):
         return not any([x.get('sequence', 0xffffffff - 1) < 0xffffffff - 1 for x in self.inputs()])
 
-    @profiler
     def estimated_size(self):
         """Return an estimated virtual tx size in vbytes.
         BIP-0141 defines 'Virtual transaction size' to be weight/4 rounded up.
@@ -1145,8 +1204,6 @@ class Transaction:
         return s, r
 
     def is_complete(self):
-        if not self.is_partial_originally:
-            return True
         s, r = self.signature_count()
         return r == s
 
@@ -1172,32 +1229,27 @@ class Transaction:
         self.raw = self.serialize()
 
     def sign_txin(self, txin_index, privkey_bytes) -> str:
-        pre_hash = Hash(bfh(self.serialize_preimage(txin_index)))
+        pre_hash = sha256d(bfh(self.serialize_preimage(txin_index)))
         privkey = ecc.ECPrivkey(privkey_bytes)
         sig = privkey.sign_transaction(pre_hash)
         sig = bh2u(sig) + '01'
         return sig
 
-    def get_outputs(self):
-        """convert pubkeys to addresses"""
-        o = []
-        for type, x, v in self.outputs():
-            if type == TYPE_ADDRESS:
-                addr = x
-            elif type == TYPE_PUBKEY:
-                # TODO do we really want this conversion? it's not really that address after all
-                addr = bitcoin.public_key_to_p2pkh(bfh(x))
+    def get_outputs_for_UI(self) -> Sequence[TxOutputForUI]:
+        outputs = []
+        for o in self.outputs():
+            if o.type == TYPE_ADDRESS:
+                addr = o.address
+            elif o.type == TYPE_PUBKEY:
+                addr = 'PUBKEY ' + o.address
             else:
-                addr = 'SCRIPT ' + x
-            o.append((addr,v))      # consider using yield (addr, v)
-        return o
+                addr = 'SCRIPT ' + o.address
+            outputs.append(TxOutputForUI(addr, o.value))      # consider using yield
+        return outputs
 
-    def get_output_addresses(self):
-        return [addr for addr, val in self.get_outputs()]
-
-
-    def has_address(self, addr):
-        return (addr in self.get_output_addresses()) or (addr in (tx.get("address") for tx in self.inputs()))
+    def has_address(self, addr: str) -> bool:
+        return (addr in (o.address for o in self.outputs())) \
+               or (addr in (txin.get("address") for txin in self.inputs()))
 
     def as_dict(self):
         if self.raw is None:
